#include <assert.h>

#ifndef _WIN32
#include <string.h>
#endif

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

#include "bot.h"

#include "bot_communicate.h"


extern int mod_id;
extern void UTIL_Radio( char * message, edict_t *pEntity);


ServerMessage * pServerMessages = NULL;

char * CopyString(const char * string)
{
	char * aCopy = (char *)malloc(sizeof(char) * (strlen(string) + 1));
	if (aCopy == NULL) exit(1);
	strcpy(aCopy, string);
	return aCopy;
}

/*************** Server Only Functions ******************/	
void PrintServerMessages()
{
	ServerMessage * pCurrent;
	while (pServerMessages != NULL)
	{
		pCurrent = pServerMessages->pNext;
		SERVER_PRINT (pCurrent->text);
		pServerMessages->pNext = pCurrent->pNext;
		if (pCurrent->pNext == pCurrent)
		{
			pServerMessages = NULL;
		}
		free(pCurrent->text);
		free(pCurrent);
	}
}

void AddServerMessage(char * message)
{
	ServerMessage * pMessage = (ServerMessage *)malloc(sizeof(ServerMessage));
	pMessage->text = CopyString(message);
	if (pServerMessages)
	{
		pMessage->pNext = pServerMessages->pNext;
		pServerMessages->pNext = pMessage;
		pServerMessages = pMessage;
	}
	else
	{
		pMessage->pNext = pMessage;
		pServerMessages = pMessage;
	}
}

/******************* Now bot Specific Stuff **************/
void BotSayText(bot_t * pBot, char * message, float delay, bool teamonly, bool radio)
{
	BotMessage * pMessage = (BotMessage *)malloc(sizeof(BotMessage));
	pMessage->text = CopyString(message);
	pMessage->time = gpGlobals->time + delay;
	pMessage->teamOnly = teamonly;
	pMessage->radio = radio;
	if (pBot->pChatMessages)
	{
		pMessage->pNext = pBot->pChatMessages->pNext;
		pBot->pChatMessages->pNext = pMessage;
		pBot->pChatMessages = pMessage;
	}
	else
	{
		pBot->pChatMessages = pMessage;
		pMessage->pNext = pMessage;
		pBot->f_chat_time = pBot->pChatMessages->time;
	}
}

void BotChat(bot_t * pBot)
{
	BotMessage  * pCurrent;
	while (pBot->pChatMessages && pBot->pChatMessages->time < gpGlobals->time)
	{
		pCurrent = pBot->pChatMessages->pNext;
		if (pCurrent->radio)
		{
			UTIL_Radio(pCurrent->text, pBot->pEdict);
		}
		else
		{
			UTIL_SayText(pCurrent->text, pBot->pEdict, pCurrent->teamOnly);
		}
		if (pCurrent == pCurrent->pNext)
		{
			pBot->pChatMessages = NULL;
		}
		else
		{
			if (pBot->pChatMessages) pBot->pChatMessages->pNext = pCurrent->pNext;
		}
		free(pCurrent->text);
		free(pCurrent);
	}
	if (pBot->pChatMessages)
	{
		pBot->f_chat_time = pBot->pChatMessages->time;
	}
}

void BotFollow(bot_t  * pBot, edict_t * pEdict)
{
	if (pEdict == NULL || pEdict == pBot->pEdict) return;
	pBot->behaviour = BEHAVE_FOLLOW;
	pBot->currentPoint = -1;
	pBot->destination = -1;
	pBot->commander = pEdict;

	if (mod_id == FLF_DLL)
	{
		char * saytext;
		if (RANDOM_LONG(0,1))
		{
			saytext = "Roger %!ROGER";
		}
		else
		{
			saytext = "You are covered %!COVERED";
		}
		BotSayText(pBot, saytext, RANDOM_FLOAT(1.0, 5.0), TRUE, TRUE);
	}
	else
	{
		BotSayText(pBot, "following", 2.0, TRUE, FALSE);
	}
}

void BotNormal(bot_t * pBot)
{
	if (pBot->behaviour != BEHAVE_NONE)
	{
		pBot->behaviour = BEHAVE_NONE;
		pBot->commander = NULL;
		BotSayText(pBot, "Roger %!ROGER", RANDOM_FLOAT(1.0, 5.0), TRUE, TRUE);
	}
}

/********** Chat Parsing Functions ***********/
void TextReceived(bot_t * pBot, char * name, char * text);

void FLF_TextReceived(bot_t * pBot, int location, int unknown, char * intext)
{
	char buffer[130];
	char text[130];
	char name[50];
	char * message;
	strcpy(text, intext+1);
	int i;
	for (i = 0; (i < 130) && (text[i] != ':') && (text[i] != '\0'); i++)
	{
		buffer[i] = text[i];
	}
	if (i <= 1) return; /* error */
	if (text[i] == '\0') return;
	if (i == 130) return;
	message = text+i+1;
	if (buffer[i-1] == ')')
	{
		/* location value */
		while ((i > 0) && (buffer[i] != '('))
		{
			i--;
		}
		if (i == 0) return;
	}
	buffer[i] = '\0'; //terminate the name;
	if (buffer[i-1] == ' ') buffer[i-1] = '\0';
	strcpy(name, buffer);
	//AddServerMessage(name);
	TextReceived(pBot, name, message);
}

void TextReceived(bot_t * pBot, char * name, char * text)
{
	char buffer[256];
	strcpy(buffer, text);
	char * words[4];
	int wordCount = 0;
	int i = 0;
	edict_t * pEdict;
	bool me = FALSE;
	for (i = 1; i <= gpGlobals->maxClients; i++)
	{
		pEdict = INDEXENT(i);
		if (pEdict)
		{
			if (strcmp(name, STRING(pEdict->v.netname)) == 0)
			{
				break;
			}
		}
	}
	if (i > gpGlobals->maxClients) pEdict = NULL;
	char * last = buffer+strlen(buffer); /* last is now the null terminator */
	while (*last != ' ' && last >= buffer) last--;
	if (*last == ' ') last++;
	if (IsAlive(pBot->pEdict) && pEdict)
	{
		if (stricmp(last, "%!COVERME\n") == 0)
		{
			Vector to = pEdict->v.origin - pBot->pEdict->v.origin;
			if (to.Length() < 500)
			{
				TraceResult tr;
				UTIL_TraceLine(pBot->pEdict->v.origin, pEdict->v.origin, ignore_monsters, pEdict, &tr);
				if (tr.flFraction >= 1.0)
				{
					BotFollow(pBot, pEdict);
					return;
				}
			}
		}
		else if (stricmp(last, "%!GOGO\n") == 0)
		{
			Vector to = pEdict->v.origin - pBot->pEdict->v.origin;
			if (to.Length() < 500)
			{
				TraceResult tr;
				UTIL_TraceLine(pBot->pEdict->v.origin, pEdict->v.origin, ignore_monsters, pEdict, &tr);
				if (tr.flFraction >= 1.0)
				{
					BotNormal(pBot);
					return;
				}
			}
		}
		else if (stricmp(last, "%!BACKUP\n") == 0)
		{
			if (pBot->behaviour == BEHAVE_NONE)
			{
				if (pBot->state != STATE_SNIPING || pBot->stateTime < gpGlobals->time)
				{
					BotFollow(pBot, pEdict);
				}
			}
		}
	}

	last = buffer;


	i=0;
	while (*last == ' ')
	{
		i++;
		last = buffer+i;
	}
	while(buffer[i] != '\0' && wordCount < 4)
	{
		if (buffer[i] == ' ' || buffer[i] == '\n')
		{
			buffer[i] = '\0';
			words[wordCount] = last;
			last = buffer+i+1;
			//sscanf(" %s ", words[wordCount], words[wordCount]);
			wordCount++;
		}
		i++;
	}

/*
	if (buffer[i] == '\0' && wordCount < 4)
	{
		words[wordCount++] = last;
	}
*/

	if (wordCount == 2)
	{
		const char * targetName = STRING(pBot->pEdict->v.netname);
		if (stricmp(words[0], STRING(pBot->pEdict->v.netname))==0)
		{
			char * command = words[1];
			if (FStrEq(command, "follow"))
			{
				if (UTIL_GetTeam(pBot->pEdict) == UTIL_GetTeam(pEdict))
				{
					BotFollow(pBot, pEdict);
				}
			}
		}
	}
}
