//functions for logging data
#ifndef _WIN32
#include <string.h>
#endif

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

#include "bot.h"
#include "bot_func.h"
#include "waypoint.h"
#include "logging.h"

extern int num_waypoints;

static bool logging = FALSE;
static bool logsLoaded = FALSE;
static bool logsInit = FALSE;

NORMAL_WAYPOINT_DATA waypoint_data[4][MAX_WAYPOINTS];
SENTRY_DATA ** sentry_data = NULL;

unsigned int totalKills = 0;
unsigned int totalDeaths = 0;
unsigned int totalMoves = 0;

float historicKills = 0;
float historicDeaths = 0;
float historicMoves = 0;

static int max_teams;

void FinishLogs(void);
extern void UTIL_BuildFileName(char *filename, char *arg1, char *arg2);

/*
* Reads the data from the map file
*/
void ReadData(void) {
	FILE* bfp;
	char map_name[50]; // = charSTRING(gpGlobals->map_name);
	char filename[150];
	int i,j;
	LOG_FILE_HEADER fileHeader;
	NORMAL_WAYPOINT_DATA pointData;
	
	InitLogs();
	strcpy(map_name, STRING(gpGlobals->mapname));
	strcat(map_name, ".odd");
	UTIL_BuildFileName(filename, "maps", map_name);
	bfp = fopen(filename, "rb");
	if (!bfp) {
		FILE * fp = fopen("log.txt", "a");
		fprintf(fp, "failed to open for reading %s \n", filename);
		fclose(fp);
		return;
	}
	
	fread(&fileHeader, sizeof(fileHeader), 1, bfp);
	
	if ( strcmp(fileHeader.mapName, STRING(gpGlobals->mapname))) { //wrong map
		fclose(bfp);
		FILE * fp = fopen("log.txt", "a");
		fprintf(fp, "Wrong map. expected %s got %s \n", STRING(gpGlobals->mapname), fileHeader.mapName);
		fclose(fp);
		return;
	}
	
	if (fileHeader.numberOfTeams != GetMaxTeam()) {
		fclose(bfp);
		FILE * fp = fopen("log.txt", "a");
		fprintf(fp, "Wrong Number of Teams. expected %d got %d \n", fileHeader.numberOfTeams, GetMaxTeam());
		fclose(fp);
		return;
	}
	
	historicMoves = fileHeader.totalMoves;
	historicKills = fileHeader.totalKills;
	historicDeaths = fileHeader.totalDeaths;
	
	for (i = 0; i < fileHeader.numberOfTeams; i++) {
		for (j = 0; j < num_waypoints; j++) {
			fread(&pointData, sizeof(pointData), 1, bfp);
			waypoint_data[i][j] = pointData;
		}
	}
	
	fclose(bfp);
}

//saves the log data and optimises the active data
void SaveData(void) {
	FILE* bfp;
	char map_name[50]; // = charSTRING(gpGlobals->map_name);
	char filename[150];
	int i;
	LOG_FILE_HEADER fileHeader;
	

	strcpy(map_name, STRING(gpGlobals->mapname));
	strcat(map_name, ".odd");
	UTIL_BuildFileName(filename, "maps", map_name);
	
	strcpy(fileHeader.mapName, STRING(gpGlobals->mapname));
	fileHeader.ver = 1;
	fileHeader.numberOfSentries = 0;
	fileHeader.numberOfTeams = max_teams;
	fileHeader.numberOfWaypoints = num_waypoints;
	
	fileHeader.totalMoves = historicMoves + totalMoves;
	fileHeader.totalKills = historicMoves + totalKills;
	fileHeader.totalDeaths = historicMoves + totalDeaths;

	bfp = fopen(filename, "wb");
	if (!bfp) {
		FILE * fp = fopen("log.txt", "a");
		fprintf(fp, "failed to open %s for writing \n", filename);
		fclose(fp);
		return;
	}
	
	fwrite(&fileHeader, sizeof(fileHeader), 1, bfp);

	//write the optimised data out to the file
	for (i = 0; i < fileHeader.numberOfTeams; i++) {
		fwrite(waypoint_data[i], sizeof(NORMAL_WAYPOINT_DATA), fileHeader.numberOfWaypoints, bfp);
	}
	fclose(bfp);

}

void InitLogs(void) {
	int i,num_teams = GetMaxTeam();
	if (logsInit) {
		FinishLogs();
	}
	max_teams = num_teams;
	for (i = 0; i < num_teams; i++) {
		memset(waypoint_data[i], 0, sizeof(NORMAL_WAYPOINT_DATA) * MAX_WAYPOINTS);
	}
	
	logsInit = TRUE;
	logging = TRUE;
	
	historicMoves = 0;
	historicKills = 0;
	historicDeaths = 0;
	
	totalMoves = 0;
	totalKills = 0;
	totalDeaths = 0;
}

void FinishLogs(void) {
	logsInit = FALSE;
	logging = FALSE;
}


void WaypointsLoaded(void) {
	InitLogs();
	ReadData();
	logsLoaded = TRUE;
	logging = TRUE;
}

void ServerDeactivated(void) {
	if (logging) {
		SaveData();
	}
	logging = FALSE;
}
	
void LogBotDie(int team, int waypoint) {
	if (waypoint < 0 || !logging) return;
	if (team < 0 || team >= max_teams) return;
	waypoint_data[team][waypoint].deaths++;
	totalDeaths++;
}

void LogBotKill(int team, int waypoint) {
	if (waypoint < 0 || !logging) return;
	if (team < 0 || team >= max_teams) return;
	waypoint_data[team][waypoint].kills++;
	totalKills++;
}

void LogBotMove(int team, int waypoint) {
	if (waypoint < 0 || !logging) return;
	if (team < 0 || team >= max_teams) return;
	waypoint_data[team][waypoint].moves++;
	totalMoves++;
}
	

float GetAverageKillDeath(void) {
	if (!logging) return 1;
	return ((double)totalKills+historicKills)/((double)totalDeaths+historicDeaths);
}

//smaller is better
float GetAverageMoveKill(void) {
	if (!logging) return 1;
	return ((double)totalMoves+historicMoves)/((double)totalKills+historicKills);
}

//smaller is better
float GetAverageDeathMove(void) {
	if (!logging) return 1;
	return ((double)totalDeaths+historicDeaths)/((double)totalMoves+historicMoves);
}

float GetKillDeath(int team, int waypoint) {
	if (!logging) return 1;
	if (team < 0 || team >= max_teams) return 1;
	if (waypoint_data[team][waypoint].kills == 0 || waypoint_data[team][waypoint].deaths == 0) return 1;
	float num =  ((double)waypoint_data[team][waypoint].kills)/((double)waypoint_data[team][waypoint].deaths);
	return num/GetAverageKillDeath();
}

float GetMoveKill(int team, int waypoint) {
	if (!logging) return 1;
	if (team < 0 || team >= max_teams) return 1; 
  if (waypoint_data[team][waypoint].kills == 0 || waypoint_data[team][waypoint].moves == 0) return 1;
  float num =  ((float)waypoint_data[team][waypoint].moves)/((float)waypoint_data[team][waypoint].kills);
	return num/GetAverageMoveKill();
}

float GetDeathMove(int team, int waypoint) {
	if (!logging) return 1;
	if (team < 0 || team >= max_teams) return 1;
	if (waypoint_data[team][waypoint].kills == 0 || waypoint_data[team][waypoint].deaths == 0) return 1;
	float num =  ((float)waypoint_data[team][waypoint].deaths)/((float)waypoint_data[team][waypoint].moves);
	return num/GetAverageDeathMove();
}

void PrintLogData(edict_t * pent) {
	char buffer[100];
	sprintf(buffer, "moves: %d, kills: %d, deaths: %d \n", totalMoves, totalKills, totalDeaths);
	ClientPrint(VARS(pent), HUD_PRINTNOTIFY, buffer );
}

void PrintPointData(edict_t *pent) {
	if (!logging) return;
	char buffer[100];
	int team;
	int point = WaypointFindNearest(pent, 80, -1);
	if (point == -1) return;
	for (team = 0; team < max_teams; team++) {
		sprintf(buffer, "Team: %d KillDeath: %f, MoveKill: %f, DeathMove: %f \n", team+1, GetKillDeath(team, point), GetMoveKill(team, point), GetDeathMove(team, point));
		ClientPrint(VARS(pent), HUD_PRINTNOTIFY, buffer);
		sprintf(buffer, " Moves: %d Kills: %d Deaths: %d \n", waypoint_data[team][point].moves, waypoint_data[team][point].kills, waypoint_data[team][point].deaths);
		ClientPrint(VARS(pent), HUD_PRINTNOTIFY, buffer);
	}
}

void LogBotSentryKill(int team, int point) {
}	

void PrintTotalData(edict_t * pent) {
	char buffer[80];
	sprintf(buffer, "moves: %f, kills: %f, deaths: %f \n", totalMoves+historicMoves, totalKills+historicKills, totalDeaths+historicDeaths);
	ClientPrint(VARS(pent), HUD_PRINTNOTIFY, buffer);
}