#ifndef _WIN32
#include <string.h>
#endif

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

#include "bot.h"
//#include "bot_func.h"
//#include "bot_client.h"
//#include "bot_weapons.h"
//#include "engine.h"
#include "waypoint.h"

#include "scripting.h"
#include "parser.h"

bool IsComment(char *string);
void RemoveComment(FILE *fp);
void ScriptError(FILE * fp, char * string);
bool AddMessage(char *message);
bool AddElement(CScriptElement *pElement, int team);
void RemoveAllMessages(void);

extern message_t *pMessages;

extern CScriptElement *pOnStart;
extern CScriptElement *pCurrentElement;
extern void UTIL_BuildFileName(char *filename, char *arg1, char *arg2);

int lineno;

bool LoadScript(void)
{
	char current_message[82];
	int current_team = -1;
	char buffer[82];
	char last_buffer[82] = "";
	char out_buffer[82];
	int level = 0; //current level in processing;
	RemoveAllMessages();
	if (!CVAR_GET_FLOAT("bot_script")) return TRUE;
	lineno = 1;
 	strcpy(out_buffer, STRING(gpGlobals->mapname));
	strcat(out_buffer, "_odd.txt");
	UTIL_BuildFileName(buffer, "maps", out_buffer);
	FILE *fp = fopen(buffer, "r");
	if (!fp) return FALSE;
	FILE *error_log = fopen("odd_bot_errors.txt", "w");
	if (!error_log) return FALSE;

	bool expecting_open = FALSE;
	bool expecting_string = FALSE;
	bool read_string = FALSE;
	bool got_expected = FALSE;
	while (!feof(fp))
	{
		*buffer = '\0';
		if (fscanf(fp,  " %[^ \t\n\r(){}] ", buffer) == 0)
		{ //it's a control character
			char c = fgetc(fp);
			//printf("Read Character: %c\n", c);
			if (c < 0) continue;
			if (c == ')')
			{
				ScriptError(error_log, "Unexpected Character: )");
				continue;
			}
			else if (c == '{')
			{
				if (expecting_open)
				{
					level++;
					expecting_open = FALSE;
				}
				else
				{
					ScriptError(error_log, "Not Expecting { yet");
				}
				continue;
			}
			else if (c == '}')
			{
				if (level > 0)
				{
					level--;
					expecting_open = FALSE;
				}
				continue;
			}
			else if (c== '(')
			{
				if (expecting_string)
				{
					if (fscanf(fp, "%80[^)])", buffer))
					{
						read_string = TRUE;
					}
					else
					{
						ScriptError(error_log, "Error reading String");
						continue;
					}
					expecting_string = FALSE;
				}
				else
				{
					ScriptError(error_log, "Not Expecting ( yet");
					continue;
				}
			}
			else if (c== '\n')
			{
				lineno++;
				continue;
			}
			else
			{
				continue;
			}
		}
		if (*buffer < 0) continue;
		if (IsComment(buffer))
		{
			RemoveComment(fp);
			lineno++;
			continue; //discard buffer
		}
		if (expecting_open)
		{
			sprintf(out_buffer, "Expecting { but got %s", buffer);
			ScriptError(error_log, out_buffer);
		}
		if (expecting_string)
		{
			sprintf(out_buffer, "Expecting ( but got %s", buffer);
			ScriptError(error_log, out_buffer);
		}

		expecting_open = FALSE;
		expecting_string = FALSE;
		got_expected = FALSE;
		if (read_string)
		{
			read_string = FALSE;
			if (level == 0)
			{
				if (strcmp(last_buffer, "on_message") == 0)
				{
					strcpy(current_message, buffer);
					AddMessage(buffer);
					expecting_open = TRUE;
					got_expected = TRUE;
				}
			}
			else if (level == 1)
			{
				if (strcmp(last_buffer, "WaypointSetType") == 0)
				{
					CScriptElement *pNewType = new CElementSetType();
					char * error_message = pNewType->Create(buffer);
					if (error_message)
					{
						ScriptError(error_log, error_message);
						delete pNewType;
					}
					else
					{
						AddElement(pNewType, current_team);
						got_expected = TRUE;
					}
				}
				else if (strcmp(last_buffer, "WaypointSetTeam") == 0)
				{
					CScriptElement *pNewTeam = new CElementSetTeam();
					char * error_message = pNewTeam->Create(buffer);
					if (error_message)
					{
						ScriptError(error_log, error_message);
						delete pNewTeam;
					}
					else
					{
						AddElement(pNewTeam, current_team);
						got_expected = TRUE;
					}
				}
				else if (strcmp(last_buffer, "ReverseCTF") == 0)
				{
					CScriptElement *pNewCTF = new CElementReverseCTF();
					char * error_message = pNewCTF->Create(buffer);
					if (error_message)
					{
						ScriptError(error_log, error_message);
						delete pNewCTF;
					}
					else
					{
						AddElement(pNewCTF, current_team);
						got_expected = TRUE;
					}
				}
				else if (strcmp(last_buffer, "SameAs") == 0)
				{
					CScriptElement *pNewCTF = new CElementSameAs();
					char * error_message = pNewCTF->Create(buffer);
					if (error_message)
					{
						ScriptError(error_log, error_message);
						delete pNewCTF;
					}
					else
					{
						AddElement(pNewCTF, current_team);
						got_expected = TRUE;
					}
				}
			}
		}
		else if (level == 0)
		{
			if (strcmp(buffer, "on_start")==0)
			{
				expecting_open = TRUE;
				got_expected = TRUE;
				current_team = -1;
				strcpy(current_message, "");
				AddMessage("");
			}
			else
			{
				expecting_open = FALSE;
			}
			if (strcmp(buffer, "on_message") == 0)
			{
				expecting_string = TRUE;
				got_expected = TRUE;
				current_team = -1;
			}
		}
		else if (level == 1)
		{
			if (strcmp(buffer, "WaypointSetType") == 0)
			{
				expecting_string = TRUE;
				got_expected = TRUE;
			}
			else if (strcmp(buffer, "WaypointSetTeam") == 0)
			{
				expecting_string = TRUE;
				got_expected = TRUE;
			}
			else if (strcmp(buffer, "ReverseCTF") == 0)
			{
				expecting_string = TRUE;
				got_expected = TRUE;
			}
			else if (strcmp(buffer, "SameAs") == 0)
			{
				expecting_string = TRUE;
				got_expected = TRUE;
			}
			else if (stricmp(buffer, "Team1") == 0)
			{
				current_team = 0;
				got_expected = TRUE;
			}
			else if (stricmp(buffer, "Team2") == 0)
			{
				current_team = 1;
				got_expected = TRUE;
			}
			else if (stricmp(buffer, "Team3") == 0)
			{
				current_team = 2;
				got_expected = TRUE;
			}
			else if (stricmp(buffer, "Team4") == 0)
			{
				current_team = 3;
				got_expected = TRUE;
			}
			else if (stricmp(buffer, "NoTeam") == 0)
			{
				current_team = -1;
				got_expected = TRUE;
			}
		}
		if (!got_expected && (*buffer != '\0'))
		{
			sprintf(out_buffer, "Unexpected Token: %s", buffer);
			ScriptError(error_log, out_buffer);
		}
		strcpy(last_buffer, buffer);
	}
	fclose(error_log);
	fclose(fp);
	return 0;
}


bool IsComment(char *buffer)
{
	if (strlen(buffer) < 2) return FALSE;
	if (*buffer == *(buffer+1) && *buffer== '/')
		return TRUE;
	return FALSE;
}

void RemoveComment(FILE *fp)
{
	while (!feof(fp) && (fgetc(fp) != '\n'));
}

void ScriptError(FILE* fp, char * string)
{
	char buffer[120];
	sprintf(buffer, "Line %d, Script Error: %s\n", lineno, string);
	fprintf(fp, "%s", buffer);
	SERVER_PRINT (buffer);
}

bool AddMessage(char *message)
{
	message_t *pNew = (message_t *)malloc(sizeof(message_t));
	if (!pNew)
	{
		return FALSE;
	}
	pNew->pNextMessage = pMessages;
	strcpy(pNew->message, message);
	pNew->pElements = NULL;
	pMessages = pNew;
	return TRUE;
}

bool AddElement(CScriptElement *pElement, int team)
{
	pElement->team = team;
	pElement->pNextElement = pMessages->pElements;
	pMessages->pElements = pElement;
	return TRUE;
}

void RemoveAllMessages(void)
{
	message_t *pMessage = pMessages;
	CScriptElement *el = NULL;
	while (pMessage != NULL)
	{
		pMessages = pMessages->pNextMessage;
		el = pMessage->pElements;
		while (el != NULL)
		{
			pMessage->pElements = pMessage->pElements->pNextElement;
			delete el;
			el = pMessage->pElements;
		}
		free(pMessage);
		pMessage = pMessages;
	}
}



