#ifndef _WIN32
#include <string.h>
#endif

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>

#include "bot.h"
#include "bot_communicate.h"
//#include "bot_func.h"
//#include "bot_client.h"
//#include "bot_weapons.h"
//#include "engine.h"
#include "waypoint.h"

#include "scripting.h"
#include "parser.h"

extern WAYPOINT waypoints[MAX_WAYPOINTS];
extern int num_waypoints;

message_t *pMessages = NULL;

CScriptElement *pOnStart = NULL;
CScriptElement *pCurrentElement = NULL;

char current_message[82];
int current_team = -1;

bool sameAsRunning = FALSE;
bot_t * currentBot = NULL;

extern tf_map_data_t tf_map;

void MessageReceived(bot_t *pBot, char *message)
{
	int team = UTIL_GetTeam(pBot->pEdict);
	bool handled = FALSE;
	message_t * pMessage = pMessages;
	CScriptElement *pElement;
	currentBot = pBot;
	while (pMessage != NULL)
	{
		//if (pMessage) pElement = pMessage->pElements; else return;
		pElement = pMessage->pElements;
		if (strcmp(message, pMessage->message) == 0)
		{
			while (pElement != NULL)
			{
				if (pElement->team == team || pElement->team == -1)
				{
					pElement->Run();
					handled = TRUE;
				}
				pElement = pElement->pNextElement;
			}
		}
		pMessage = pMessage->pNextMessage;
	}
	if (CVAR_GET_FLOAT("bot_message_debug"))
	{
		char buffer[256];
		sprintf(buffer, "Message Received by %s: '%s' %s\n", STRING(pBot->pEdict->v.netname), message, handled?"Handled":"Not Handled");
		AddServerMessage(buffer);
	}
	currentBot = NULL;
}

void ScriptOnStart(void)
{
	message_t * pMessage = pMessages;
	CScriptElement *pElement;
	while (pMessage != NULL)
	{
		pElement = pMessage->pElements;
		if (strcmp(pMessage->message, "") == 0)
		{
			while (pElement != NULL)
			{
				pElement->Run();
				pElement = pElement->pNextElement;
			}
		}
		pMessage = pMessage->pNextMessage;
	}
}
/********************** Base Class *****************/
CScriptElement::CScriptElement()
{
	pNextElement = NULL;
}

void CScriptElement::Run(void)
{
}


char * CScriptElement::Create(char * args)
{
	return "Error, unitialised Script Element";
}

/******************** Set Flag *********************/

CElementSetType::CElementSetType()
{
	waypoint = 0;
	flag = 0;
	isnot = FALSE;
}

void CElementSetType::Run(void)
{
	if (this->isnot)
	{
		waypoints[this->waypoint].flags &= ~(this->flag);
	}
	else
	{
		waypoints[this->waypoint].flags |= this->flag;
	}
}

char * CElementSetType::Create(char *args)
{
	char waypoint[80];
	char flags[80];
	if (sscanf(args, " %[^ ,] %*[ ,] %s", waypoint, flags) == 2)
	{
		printf("Success!!!\n");
		printf("waypoint: %s, flags: %s\n", waypoint, flags);
		if (*flags == '!')
		{
			char temp[80];
			strcpy(temp, flags);
			sscanf(temp, " ! %s", flags);
			this->isnot = TRUE;
		}
		this->waypoint = atoi(waypoint);
		if (this->waypoint <= 0)
		{
			return "Error, illegal waypoint";
		}
		this->waypoint --;
		if (this->waypoint >= num_waypoints)
		{
			return "Illegal waypoint";
		}
		if (stricmp(flags, "flag") == 0)
		{
			this->flag = W_FL_FLAG;
		}
		else if (stricmp(flags, "flag_goal")==0)
		{
			this->flag = W_FL_FLAG_GOAL;
		}
		else if (stricmp(flags, "detpack")==0)
		{
			this->flag = W_FL_DETPACK;
		}
		else if (stricmp(flags, "guard") == 0)
		{
			this->flag = W_FL_GUARD;
		}
		else if (stricmp(flags, "sniper") == 0)
		{
			this->flag = W_FL_SNIPER;
		}
		else
		{
			return "Uncecognised waypoint type";
		}
		return NULL;
	}
	return "Error Initing CElementSetFlag";
}

/************* Set Team **********************************/
CElementSetTeam::CElementSetTeam()
{
	waypoint = 0;
	toteam = 0;
}

void CElementSetTeam::Run(void)
{
	if (toteam == 0)
	{
		waypoints[waypoint].flags &= ~W_FL_TEAM_SPECIFIC;
		waypoints[waypoint].flags &= ~W_FL_TEAM;
	}
	else
	{
		int num = toteam-1;
		waypoints[waypoint].flags |= W_FL_TEAM_SPECIFIC;
		waypoints[waypoint].flags |= num;
	}
}

char * CElementSetTeam::Create(char *args)
{
	char waypoint[80];
	int newteam= -1;
	if (sscanf(args, " %[^ ,] %*[ ,] %d", waypoint, &newteam) == 2)
	{
		printf("Team Success!!!\n");
		printf("waypoint: %s, team: %d\n", waypoint, newteam);
		this->waypoint = atoi(waypoint);
		if (this->waypoint <= 0)
		{
			return "Error, illegal waypoint";
		}
		this->waypoint --;
		if (this->waypoint >= num_waypoints)
		{
			return "Illegal waypoint";
		}
		if (newteam < 0)
		{
			return "Team must be greater than -1";
		}
		else if (newteam > 4)
		{
			return "Team must be less than 5";
		}
		this->toteam = newteam;
		return NULL;
	}
	return "Error Initing CElementSetFlag";
}

/*********** CElementReverseCTF ******************/
CElementReverseCTF::CElementReverseCTF()
{
	this->on = FALSE;
}

void CElementReverseCTF::Run()
{
	tf_map.reverseCTF = this->on;
}

char * CElementReverseCTF::Create(char * args)
{
	if (strcmp(args, "true")==0)
	{
		on = TRUE;
	}
	else if (strcmp(args, "false")==0)
	{
		on = FALSE;
	}
	else
	{
		return "ReverseCTF argument must be true or false";
	}
	return NULL;
}

/****************** CElementSameAs ******************/
CElementSameAs::CElementSameAs()
{
	this->string = NULL;
}

CElementSameAs::~CElementSameAs()
{
	if (string)
	{
		free(string);
	}
}

char * CElementSameAs::Create(char * args)
{
	if (strlen(args) == 0) return "SameAs must have an argument";
	this->string = (char *)malloc(sizeof(char) * (strlen(args)+1));
	strcpy(string, args);
	return NULL;
}

void CElementSameAs::Run()
{
	bot_t * oldBot = NULL;
	if (sameAsRunning) return;
	if (currentBot)
	{
		sameAsRunning = TRUE;
		oldBot = currentBot;
		MessageReceived(currentBot, string);
		currentBot = oldBot;
		sameAsRunning = FALSE;
	}
}

		